/*
 * Copyright (c) 2019 Tavish Naruka <tavishnaruka@gmail.com>
 *
 * SPDX-License-Identifier: Apache-2.0
 */

/* Sample which uses the filesystem API and SDHC driver */

#include <zephyr/zephyr.h>
#include <zephyr/device.h>
#include <zephyr/storage/disk_access.h>
#include <zephyr/logging/log.h>
#include <zephyr/fs/fs.h>
#include <ff.h>

#include <soc.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/sdhc.h>
#include <hal/nrf_gpio.h>
#include <zephyr/pm/pm.h>
#include <zephyr/pm/device.h>
#include <zephyr/pm/policy.h>

#define SPI_NODE						DT_NODELABEL(spi1)
#define IO_V_ENABLE_NODE  				DT_NODELABEL(io_v_en)
#define SCK_NODE  				DT_NODELABEL(sck)
#define MISO_NODE  				DT_NODELABEL(miso)
#define MOSI_NODE  				DT_NODELABEL(mosi)
#define CS_NODE					DT_NODELABEL(cs)
#define CARD_NODE				DT_NODELABEL(sdhc0)
#if !(DT_NODE_HAS_STATUS(SPI_NODE, okay) && DT_NODE_HAS_STATUS(IO_V_ENABLE_NODE, okay) \
&& DT_NODE_HAS_STATUS(SCK_NODE, okay) && DT_NODE_HAS_STATUS(MISO_NODE, okay) && DT_NODE_HAS_STATUS(MOSI_NODE, okay) && DT_NODE_HAS_STATUS(CS_NODE, okay) && DT_NODE_HAS_STATUS(CARD_NODE, okay))
#error "Selected board is not supported"
#else
const struct device* spi_dev 			= DEVICE_DT_GET(SPI_NODE);
static struct gpio_dt_spec io_v_enable 	= GPIO_DT_SPEC_GET(IO_V_ENABLE_NODE, gpios);
static struct gpio_dt_spec sck_en 	= GPIO_DT_SPEC_GET(SCK_NODE, gpios);
static struct gpio_dt_spec miso_en 	= GPIO_DT_SPEC_GET(MISO_NODE, gpios);
static struct gpio_dt_spec mosi_en 	= GPIO_DT_SPEC_GET(MOSI_NODE, gpios);
static struct gpio_dt_spec cs_en 	= GPIO_DT_SPEC_GET(CS_NODE, gpios);
const struct device* sdhc_dev	= DEVICE_DT_GET(CARD_NODE);
#endif

LOG_MODULE_REGISTER(main);

static int lsdir(const char *path);

static FATFS fat_fs;
/* mounting info */
static struct fs_mount_t mp = {
	.type = FS_FATFS,
	.fs_data = &fat_fs,
};

/*
*  Note the fatfs library is able to mount only strings inside _VOLUME_STRS
*  in ffconf.h
*/
static const char *disk_mount_pt = "/SD:";

extern int disk_access_unregister(struct disk_info *disk);
extern struct disk_info *disk_access_get_di(const char *name);

void mount_and_unmount(){
	// Provde power to sd card
	gpio_pin_set_dt(&io_v_enable, 1);
	gpio_pin_set_dt(&cs_en, 1);
	k_msleep(1000);

	int present = sdhc_card_present(sdhc_dev);
	if(present == 1){
		printk("Card present.\n");
	} else if (present == 0){
		printk("Card not present.\n");
	} else {
		printk("Failed detecting presentness.\n");
	}

	/* raw disk i/o */
	static const char *disk_pdrv = "SD";
	do {
		uint64_t memory_size_mb;
		uint32_t block_count;
		uint32_t block_size;

		if (disk_access_init(disk_pdrv) != 0) {
			LOG_ERR("Storage init ERROR!");
			break;
		}

		if (disk_access_ioctl(disk_pdrv,
				DISK_IOCTL_GET_SECTOR_COUNT, &block_count)) {
			LOG_ERR("Unable to get sector count");
			break;
		}
		LOG_INF("Block count %u", block_count);

		if (disk_access_ioctl(disk_pdrv,
				DISK_IOCTL_GET_SECTOR_SIZE, &block_size)) {
			LOG_ERR("Unable to get sector size");
			break;
		}
		printk("Sector size %u\n", block_size);

		memory_size_mb = (uint64_t)block_count * block_size;
		printk("Memory Size(MB) %u\n", (uint32_t)(memory_size_mb >> 20));
	} while (0);

	mp.mnt_point = disk_mount_pt;

	int res = fs_mount(&mp);

	if (res == FR_OK) {
		printk("Disk mounted.\n");
		lsdir(disk_mount_pt);
	} else {
		printk("Error mounting disk.\n");
	}

	k_msleep(2000);

	fs_unmount(&mp);

	// Doesn't work.
	// k_msleep(2000);
	// disk_access_unregister(disk_access_get_di(disk_pdrv));
	// Doesn't work.

	// static struct sdhc_io sdhc_io_inst = {SD_CLOCK_25MHZ, SDHC_BUSMODE_PUSHPULL, SDHC_POWER_OFF, SDHC_BUS_WIDTH1BIT, SDHC_TIMING_HS, SD_DRIVER_TYPE_A, SD_VOL_3_3_V};
	// sdhc_set_io(sdhc_dev, &sdhc_io_inst);

	k_msleep(2000);
	// Commenting line underneath works. This line removes power from SD-card, which effectively is the same as removing it. It also results in exactly the same
	// behaviour as ejecting and inserting it.
	// gpio_pin_set_dt(&io_v_enable, 0);
	pm_device_action_run(spi_dev, PM_DEVICE_ACTION_SUSPEND);
	k_msleep(2000);
	gpio_pin_set_dt(&cs_en, 1);
}

void main(void)
{
	gpio_pin_configure_dt(&io_v_enable, GPIO_OUTPUT_ACTIVE);
	gpio_pin_configure_dt(&cs_en, GPIO_OUTPUT_ACTIVE);

	while(true){
		mount_and_unmount();
		k_msleep(10000);
	}
}

static int lsdir(const char *path)
{
	int res;
	struct fs_dir_t dirp;
	static struct fs_dirent entry;

	fs_dir_t_init(&dirp);

	/* Verify fs_opendir() */
	res = fs_opendir(&dirp, path);
	if (res) {
		printk("Error opening dir %s [%d]\n", path, res);
		return res;
	}

	printk("\nListing dir %s ...\n", path);
	for (;;) {
		/* Verify fs_readdir() */
		res = fs_readdir(&dirp, &entry);

		/* entry.name[0] == 0 means end-of-dir */
		if (res || entry.name[0] == 0) {
			break;
		}

		if (entry.type == FS_DIR_ENTRY_DIR) {
			printk("[DIR ] %s\n", entry.name);
		} else {
			printk("[FILE] %s (size = %zu)\n",
				entry.name, entry.size);
		}
	}

	/* Verify fs_closedir() */
	fs_closedir(&dirp);

	return res;
}
